// license:BSD-3-Clause
// copyright-holders:David Haywood
/*

GX840 PWB351656 PCB
with GX840 PWB451716 sub-board for some ROMs

the PCB has a spaghetti of wiremods.

sprite chip appars to be MN53060KGB

2x K051316 for ROZ layers

4x8 DSW


todo / to verify:
game hangs (loops one scene forever) ofen, presumably communication breakdown (or did? might depend on an input bit in the coin area, or a dipswitch??)
game hangs (loops one scene forever) if you die from hitting a regular enemy at the same time a boss dies?
boss can't kill you directly(?) maybe due to prototype status??
palette select not fully understood, all we know is that front roz should never use pal 0x03
verify layer alignment
sometimes a stray bad explosion sprite after you respawn
lives counter is corrupt for >4 lives (almost certainly an original game bug)
game has no sprites when you reach level 4 and becomes unplayable?


*/

#include "emu.h"
#include "cpu/z80/z80.h"
#include "sound/sn76496.h"
#include "video/k051316.h"

#define XTAL1 XTAL_24MHz

class divebomb_state : public driver_device
{
public:
	divebomb_state(const machine_config &mconfig, device_type type, const char *tag)
		: driver_device(mconfig, type, tag),
		m_spritecpu(*this, "spritecpu"),
		m_fgcpu(*this, "fgcpu"),
		m_rozcpucpu(*this, "rozcpu"),
		m_bank1(*this, "bank1"),
		m_fgram(*this, "fgram"),
		m_spriteram(*this, "spriteram"),
		m_gfxdecode(*this, "gfxdecode"),
		m_palette(*this, "palette"),
		m_k051316_1(*this, "k051316_1"),
		m_k051316_2(*this, "k051316_2")
	{ }

	virtual void machine_start() override;
	virtual void machine_reset() override;
	virtual void video_start() override;
	UINT32 screen_update_divebomb(screen_device &screen, bitmap_ind16 &bitmap, const rectangle &cliprect);
	required_device<cpu_device> m_spritecpu;
	required_device<cpu_device> m_fgcpu;
	required_device<cpu_device> m_rozcpucpu;
	required_memory_bank m_bank1;
	required_shared_ptr<UINT8> m_fgram;
	required_shared_ptr<UINT8> m_spriteram;
	required_device<gfxdecode_device> m_gfxdecode;
	required_device<palette_device> m_palette;
	required_device<k051316_device> m_k051316_1;
	required_device<k051316_device> m_k051316_2;

	void decode_proms(UINT8* rgn, int size, int index, int inv);
	DECLARE_PALETTE_INIT(divebomb);

	tilemap_t *m_fg_tilemap;
	DECLARE_WRITE8_MEMBER(fgram_w);
	TILE_GET_INFO_MEMBER(get_fg_tile_info);

	K051316_CB_MEMBER(zoom_callback_1);
	K051316_CB_MEMBER(zoom_callback_2);

	int roz1_enable;
	int roz2_enable;
	int roz1_wrap;
	int roz2_wrap;
	int to_spritecpu;
	int to_rozcpu;
	int has_fromsprite;
	int has_fromroz;
	int fromsprite;
	int fromroz;
	int port50_data;

	void draw_sprites(bitmap_ind16 &bitmap, const rectangle &cliprect);


	


	DECLARE_WRITE8_MEMBER(rozcpu_bank_w)
	{
		int bank = BITSWAP8(data, 4, 5, 6, 7, 3, 2, 1, 0);
		bank = bank >> 4;
		m_bank1->set_entry(bank);

		if (data & 0x0f)
			printf("rozcpu_bank_w %02x\n", data);
	}




	DECLARE_WRITE8_MEMBER(rozcpu_wrap1_enable_w)
	{
	//	if ((data != 0x00) && (data != 0x01))
	//		printf("rozcpu_wrap1_enable_w %02x\n", data);
		roz1_wrap = data & 1;

	}

	DECLARE_WRITE8_MEMBER(rozcpu_enable1_w)
	{
	//	if ((data != 0x00) && (data != 0x01))
	//	printf("rozcpu_enable1_w %02x\n", data);

		roz1_enable = data & 1;
	}

	DECLARE_WRITE8_MEMBER(rozcpu_enable2_w)
	{
	//	if ((data != 0x00) && (data != 0x01))
	//	printf("rozcpu_enable2_w %02x\n", data);

		roz2_enable = data & 1;
	}

	DECLARE_WRITE8_MEMBER(rozcpu_wrap2_enable_w)
	{
	//	if ((data != 0x00) && (data != 0x01))
	//		printf("rozcpu_wrap2_enable_w %02x\n", data);
		roz2_wrap = data & 1;
	}



	void update_irqs()
	{
		if (has_fromsprite || has_fromroz)
			m_fgcpu->set_input_line(INPUT_LINE_IRQ0, ASSERT_LINE);
		else
			m_fgcpu->set_input_line(INPUT_LINE_IRQ0, CLEAR_LINE);
	}

	DECLARE_READ8_MEMBER(fgcpu_port20_r)
	{
		has_fromsprite = 0;
		update_irqs();
		return fromsprite;//rand();
	}

	DECLARE_WRITE8_MEMBER(fgcpu_port20_w)
	{
		m_spritecpu->set_input_line(INPUT_LINE_IRQ0, ASSERT_LINE);
		to_spritecpu = data;
	}

	DECLARE_READ8_MEMBER(fgcpu_port10_r)
	{
		has_fromroz = 0;
		update_irqs();
		return fromroz;//rand();
	}

	DECLARE_WRITE8_MEMBER(fgcpu_port10_w)
	{
		m_rozcpucpu->set_input_line(INPUT_LINE_IRQ0, ASSERT_LINE);
		to_rozcpu = data;
	}

	DECLARE_READ8_MEMBER(spritecpu_port80_r)
	{
		UINT8 ret = to_spritecpu;
		to_spritecpu = 0x00;
		m_spritecpu->set_input_line(INPUT_LINE_IRQ0, CLEAR_LINE);
		return ret;
	}

	DECLARE_WRITE8_MEMBER(spritecpu_port80_w)
	{
		fromsprite = data;
		has_fromsprite = 1;
		update_irqs();
	}

	DECLARE_WRITE8_MEMBER(spritecpu_port00_w)
	{
		printf("spritecpu_port00_w %02x\n", data);
	}
	

	DECLARE_READ8_MEMBER(rozcpu_port40_r)
	{
		UINT8 ret = to_rozcpu;
		to_rozcpu = 0x00;
		m_rozcpucpu->set_input_line(INPUT_LINE_IRQ0, CLEAR_LINE);
		return ret;
	}

	DECLARE_WRITE8_MEMBER(rozcpu_port40_w)
	{
		fromroz = data;
		has_fromroz = 1;
		update_irqs();
	}


	DECLARE_WRITE8_MEMBER(rozcpu_port50_w)
	{
		// initially I thought the lower 4-bits were to do colour cycling on the front-most
		// roz tilemap, as the values written during gameplay are 00, 01, and 02 - the front layer has 3 valid palettes.
		// however, during a boss fight a value of 0x03 is written, which is invalid for the fg layer.
		//
		// the background layer however has 4 valid palettes, the 4th one being a darkened version, this makes more sense.
		//
		// the upper 4 bits might be for the front layer instead, one bit toggles when you hit a boss causing a palette effect.
		port50_data = data;

		m_k051316_2->mark_tmap_dirty();
		m_k051316_1->mark_tmap_dirty();

		if ( data & 0xfc ) printf("rozcpu_port50_w %02x\n", data);
	}







	DECLARE_READ8_MEMBER(fgcpu_port37_r)
	{
		// the lowest 2 bits are checked in the irq
		// probably checks which cpus have a command?
		// as result determines if to read/write por 0x10 or 0x20
		/*
		
		0103: DB 37         in   a,($37)
		0105: E6 02         and  $02
		0107: C2 1C 01      jp   nz,$011C
		
		010A: DB 10         in   a,($10)
		010C: 4F            ld   c,a
		010D: 3A 66 E0      ld   a,($E066)
		0110: B9            cp   c
		0111: C2 44 01      jp   nz,$0144
		0114: CB 8E         res  1,(hl)
		0116: E1            pop  hl
		0117: C1            pop  bc
		0118: F1            pop  af
		0119: FB            ei
		011A: ED 4D         reti

		011C: DB 20         in   a,($20)
		011E: 4F            ld   c,a
		011F: 3A 67 E0      ld   a,($E067)
		0122: B9            cp   c
		0123: C2 03 02      jp   nz,$0203
		0126: CB 86         res  0,(hl)
		0128: E1            pop  hl
		0129: C1            pop  bc
		012A: F1            pop  af
		012B: FB            ei
		012C: ED 4D         reti

		...

		012E: DB 37         in   a,($37)
		0130: E6 02         and  $02
		0132: C2 05 02      jp   nz,$0205
		0135: DB 37         in   a,($37)
		0137: E6 01         and  $01
		0139: 20 06         jr   nz,$0141
		013B: E1            pop  hl
		013C: C1            pop  bc
		013D: F1            pop  af
		013E: FB            ei
		013F: ED 4D         reti
			
		*/
		int ret = 0;
		if (has_fromroz) ret |= 1;
		if (has_fromsprite) ret |= 2;

		return ret;
		//return rand() & 0x03;
	}


};

TILE_GET_INFO_MEMBER(divebomb_state::get_fg_tile_info)
{
	int code = m_fgram[tile_index+0x000];
	int attr = m_fgram[tile_index+0x400];
	
	code |= (attr & 0x3) << 8;

	int colour = attr >> 4;

	SET_TILE_INFO_MEMBER(0,
			code,
			colour,
			0);
}

WRITE8_MEMBER(divebomb_state::fgram_w)
{
	m_fgram[offset] = data;
	m_fg_tilemap->mark_tile_dirty(offset&0x3ff);
}

void divebomb_state::draw_sprites( bitmap_ind16 &bitmap, const rectangle &cliprect )
{
	UINT8 *spriteram = m_spriteram;
	int i;

	for (i = 0; i < m_spriteram.bytes(); i += 4)
	{
		int sy = spriteram[i + 3];
		int sx = spriteram[i + 0];

		int code = spriteram[i + 2];
		int attr = spriteram[i + 1];

		code += (attr & 0x0f) << 8;

		int colour = attr >> 4;

		m_gfxdecode->gfx(1)->transpen(bitmap,cliprect, code, colour, 0, 0,
			sx, sy,
			0);

		m_gfxdecode->gfx(1)->transpen(bitmap,cliprect, code, colour, 0, 0,
			sx, sy-256,
			0);
	}
}

void divebomb_state::video_start()
{
	m_fg_tilemap = &machine().tilemap().create(m_gfxdecode, tilemap_get_info_delegate(FUNC(divebomb_state::get_fg_tile_info),this), TILEMAP_SCAN_ROWS, 8, 8, 32, 32);
	m_fg_tilemap->set_transparent_pen(0);
	m_fg_tilemap->set_scrolly(0, 16);
}


UINT32 divebomb_state::screen_update_divebomb(screen_device &screen, bitmap_ind16 &bitmap, const rectangle &cliprect)
{
	bitmap.fill(0, cliprect);
	
	m_k051316_2->wraparound_enable(!roz1_wrap);
	m_k051316_1->wraparound_enable(!roz2_wrap);
	
	if (!roz2_enable) m_k051316_2->zoom_draw(screen, bitmap, cliprect, 0, 0);
	if (!roz1_enable) m_k051316_1->zoom_draw(screen, bitmap, cliprect, 0, 0);



	draw_sprites(bitmap, cliprect);

	m_fg_tilemap->draw(screen, bitmap, cliprect, 0, 0);
	return 0;
}

// presumably this is the main cpu, handles the fg layer, audio and comms with the others?

static ADDRESS_MAP_START( divebomb_fgcpu_map, AS_PROGRAM, 8, divebomb_state )
	AM_RANGE(0x0000, 0x7fff) AM_ROM
	AM_RANGE(0xc000, 0xc7ff) AM_RAM_WRITE(fgram_w) AM_SHARE("fgram")
	AM_RANGE(0xc800, 0xcfff) AM_RAM
	AM_RANGE(0xe000, 0xffff) AM_RAM
ADDRESS_MAP_END

static ADDRESS_MAP_START( divebomb_fgcpu_iomap, AS_IO, 8, divebomb_state )
	ADDRESS_MAP_GLOBAL_MASK(0xff)
	AM_RANGE(0x00, 0x00) AM_DEVWRITE("sn0", sn76489_device, write)
	AM_RANGE(0x01, 0x01) AM_DEVWRITE("sn1", sn76489_device, write)
	AM_RANGE(0x02, 0x02) AM_DEVWRITE("sn2", sn76489_device, write)
	AM_RANGE(0x03, 0x03) AM_DEVWRITE("sn3", sn76489_device, write)
	AM_RANGE(0x04, 0x04) AM_DEVWRITE("sn4", sn76489_device, write)
	AM_RANGE(0x05, 0x05) AM_DEVWRITE("sn5", sn76489_device, write)

	AM_RANGE(0x10, 0x10) AM_READWRITE(fgcpu_port10_r, fgcpu_port10_w)
	AM_RANGE(0x20, 0x20) AM_READWRITE(fgcpu_port20_r, fgcpu_port20_w)

	AM_RANGE(0x30, 0x30) AM_READ_PORT("IN0") 
	AM_RANGE(0x31, 0x31) AM_READ_PORT("IN1") 

	AM_RANGE(0x32, 0x32) AM_READ_PORT("DSW0") 
	AM_RANGE(0x33, 0x33) AM_READ_PORT("DSW1") 
	// 34/35 aren't read but pcb has 4 banks of dips populated, so logically they would map here even if unused
	AM_RANGE(0x34, 0x34) AM_READ_PORT("DSW2") 
	AM_RANGE(0x35, 0x35) AM_READ_PORT("DSW3") 


	AM_RANGE(0x36, 0x36) AM_READ_PORT("SYSTEM") 
	AM_RANGE(0x37, 0x37) AM_READ(fgcpu_port37_r)
ADDRESS_MAP_END

// sprites only?
static ADDRESS_MAP_START( divebomb_spritecpu_map, AS_PROGRAM, 8, divebomb_state )
	AM_RANGE(0x0000, 0x7fff) AM_ROM
	AM_RANGE(0xc000, 0xcfff) AM_RAM AM_SHARE("spriteram") // 0x200 chunks, might be selectable?
	AM_RANGE(0xd000, 0xdfff) AM_RAM
	AM_RANGE(0xe000, 0xffff) AM_RAM
ADDRESS_MAP_END

static ADDRESS_MAP_START( divebomb_spritecpu_iomap, AS_IO, 8, divebomb_state )
	ADDRESS_MAP_GLOBAL_MASK(0xff)
	AM_RANGE(0x00, 0x00) AM_WRITE(spritecpu_port00_w)
	AM_RANGE(0x80, 0x80) AM_READWRITE(spritecpu_port80_r, spritecpu_port80_w)
ADDRESS_MAP_END


// roz only?
static ADDRESS_MAP_START( divebomb_rozcpu_map, AS_PROGRAM, 8, divebomb_state )
	AM_RANGE(0x0000, 0x7fff) AM_ROM
	AM_RANGE(0x8000, 0xbfff) AM_ROMBANK("bank1") 
	AM_RANGE(0xc000, 0xc7ff) AM_RAM AM_DEVREADWRITE("k051316_1", k051316_device, read, write)
	AM_RANGE(0xd000, 0xd7ff) AM_RAM AM_DEVREADWRITE("k051316_2", k051316_device, read, write)
	AM_RANGE(0xe000, 0xffff) AM_RAM
ADDRESS_MAP_END

static ADDRESS_MAP_START( divebomb_rozcpu_iomap, AS_IO, 8, divebomb_state )
	ADDRESS_MAP_GLOBAL_MASK(0xff)
	AM_RANGE(0x00, 0x00) AM_WRITE( rozcpu_bank_w )

	AM_RANGE(0x10, 0x10) AM_WRITE(rozcpu_wrap1_enable_w) // 0/1
	AM_RANGE(0x12, 0x12) AM_WRITE(rozcpu_enable1_w) // 0/1, probably roz enable or wrap enable.
	AM_RANGE(0x13, 0x13) AM_WRITE(rozcpu_enable2_w) // ^
	AM_RANGE(0x14, 0x14) AM_WRITE(rozcpu_wrap2_enable_w) // ^

	AM_RANGE(0x20, 0x2b) AM_DEVWRITE("k051316_1", k051316_device, ctrl_w) 
	AM_RANGE(0x2c, 0x2e) AM_WRITENOP // still k051316 regs?

	AM_RANGE(0x30, 0x3b) AM_DEVWRITE("k051316_2", k051316_device, ctrl_w) 
	AM_RANGE(0x3c, 0x3e) AM_WRITENOP // still k051316 regs?

	AM_RANGE(0x40, 0x40) AM_READWRITE(rozcpu_port40_r, rozcpu_port40_w) // comms?
	AM_RANGE(0x50, 0x50) AM_WRITE(rozcpu_port50_w) // sequences of 0 then 1, then 2
ADDRESS_MAP_END






static INPUT_PORTS_START( divebomb )
	PORT_START("DSW0")
	PORT_DIPNAME( 0x01, 0x01, "DSW0" ) // if the first 4 are on it engages freeplay
	PORT_DIPSETTING(    0x01, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x02, 0x02, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x02, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x04, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x08, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x08, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x20, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START("DSW1")
	PORT_DIPNAME( 0x03, 0x02, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x03, "2" )
	PORT_DIPSETTING(    0x02, "4" )
	PORT_DIPSETTING(    0x01, "7" ) // shows corrupt numbers
	PORT_DIPSETTING(    0x00, "8" ) // shows 0 then corrupt numbers
	PORT_DIPNAME( 0x04, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x08, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START("DSW2")
	PORT_DIPNAME( 0x01, 0x00, "DSW2" )
	PORT_DIPSETTING(    0x01, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x02, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x02, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x08, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START("DSW3")
	PORT_DIPNAME( 0x01, 0x01, "DSW3" )
	PORT_DIPSETTING(    0x01, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x02, 0x02, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x02, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x04, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x08, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x08, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x20, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START("IN0")
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP ) PORT_8WAY // dive
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT ) PORT_8WAY
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN ) PORT_8WAY // lift
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT ) PORT_8WAY
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_BUTTON2 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_BUTTON3 ) // shows 'surface scanner'

	PORT_START("IN1")
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP ) PORT_PLAYER(2) PORT_8WAY // dive
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT ) PORT_PLAYER(2) PORT_8WAY
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN ) PORT_PLAYER(2) PORT_8WAY // lift
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT ) PORT_PLAYER(2) PORT_8WAY
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 ) PORT_PLAYER(2)
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_BUTTON2 ) PORT_PLAYER(2)
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_BUTTON3 ) PORT_PLAYER(2) // shows 'surface scanner'

	PORT_START("SYSTEM")
	PORT_DIPNAME( 0x01, 0x00, "0" )
	PORT_DIPSETTING(    0x01, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x02, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x02, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

INPUT_PORTS_END



static const gfx_layout tiles8x8_layout =
{
	8,8,
	RGN_FRAC(1,1),
	2,
	{ 8,0 },
	{ 0, 1, 2, 3, 4, 5, 6, 7 },
	{ 0*16, 1*16, 2*16, 3*16, 4*16, 5*16, 6*16, 7*16 },
	16*8
};

static const gfx_layout tiles16x16_layout =
{
	16,16,
	RGN_FRAC(1,1),
	4,
	{ 24,16,8,0 },
	{ 0, 1, 2, 3, 4, 5, 6, 7, 32,33,34,35,36,37,38,39 },
	{ 0*64, 1*64, 2*64, 3*64, 4*64, 5*64, 6*64, 7*64, 8*64,9*64,10*64,11*64,12*64,13*64,14*64,15*64 },
	64*16
};



static GFXDECODE_START( divebomb )
	GFXDECODE_ENTRY( "gfx1", 0, tiles8x8_layout, 0x400+0x400, 16 )
	GFXDECODE_ENTRY( "gfx2", 0, tiles16x16_layout, 0x400+0x400+0x400, 16 )
GFXDECODE_END


void divebomb_state::decode_proms(UINT8* rgn, int size, int index, int inv)
{
	
	int i;

	for (i = 0;i < size;i++)
	{
		int r,g,b;
		r= (rgn[i+size*2]) & 0x0f;
		r = r << 4;
		g = (rgn[i+size]) & 0x0f;
		g = g << 4;	
		b = (rgn[i]) & 0x0f;
		b = b << 4;

		if (!inv) m_palette->set_pen_color(index+i,rgb_t(r,g,b));
		else  m_palette->set_pen_color(index+(i^0xff),rgb_t(r,g,b));

	}
}

PALETTE_INIT_MEMBER(divebomb_state, divebomb)
{

	decode_proms(memregion("spr_proms")->base(), 0x100, 0x400 + 0x400 + 0x400, 0); // sprites
	decode_proms(memregion("fg_proms")->base(), 0x400, 0x400 + 0x400, 0); // text

	decode_proms(memregion("k051316_1_pr")->base(), 0x400, 0, 1); // front roz layer

	decode_proms(memregion("k051316_2_pr")->base(), 0x400, 0x400, 1); // back roz layer


}

void divebomb_state::machine_start()
{
	m_bank1->configure_entries(0, 16, memregion("rozcpudata")->base(), 0x4000);

	save_item(NAME(roz1_enable));
	save_item(NAME(roz2_enable));
	save_item(NAME(roz1_wrap));
	save_item(NAME(roz2_wrap));
	save_item(NAME(to_spritecpu));
	save_item(NAME(to_rozcpu));

	save_item(NAME(has_fromsprite));
	save_item(NAME(has_fromroz));

	save_item(NAME(fromsprite));
	save_item(NAME(fromroz));

	save_item(NAME(port50_data));
}

void divebomb_state::machine_reset()
{
}

K051316_CB_MEMBER(divebomb_state::zoom_callback_1)
{
	int colourbase = 0;
	*code |= ((*color & 0x03) << 8);

//	if (*color & 0xfc) printf("zoom_callback_1 color & 0x30\n");

	// bits 0x40 and 0x80 get set, but sometimes the corner of a firing turret get set to 0xc0, which isn't a valid palette
	// ** how should we use the alt palette banks, 0,1,2 are minor palette cycles of eac other - guesses below

	// we can't use this for colour cycling as it gets set to 0x03 during boss battle.
	//	*color += port50_data & 3;

	// as only 3 roms are populated for this layer maybe each palette relates to one of the ROMs?
	// meaning the tile bits are reused as palette bits intead.
	//*color = colourbase + (*color & 0x03);

	// actually most logical is that the upper 4 bits are used instead, they toggle on a boss hit
	*color = colourbase;
	*color += (port50_data>>4) & 3;
}

K051316_CB_MEMBER(divebomb_state::zoom_callback_2)
{
	int colourbase = 4;
	*code |= ((*color & 0x03) << 8);
	*color = colourbase;
	*color += port50_data & 3;
}


static MACHINE_CONFIG_START( divebomb, divebomb_state )

	MCFG_CPU_ADD("fgcpu", Z80,XTAL1/4)         /* ? MHz */
	MCFG_CPU_PROGRAM_MAP(divebomb_fgcpu_map)
	MCFG_CPU_IO_MAP(divebomb_fgcpu_iomap)
	MCFG_CPU_VBLANK_INT_DRIVER("screen", divebomb_state,  nmi_line_pulse)

	/* basic machine hardware */
	MCFG_CPU_ADD("spritecpu", Z80,XTAL1/4)         /* ? MHz */
	MCFG_CPU_PROGRAM_MAP(divebomb_spritecpu_map)
	MCFG_CPU_IO_MAP(divebomb_spritecpu_iomap)
	MCFG_CPU_VBLANK_INT_DRIVER("screen", divebomb_state,  nmi_line_pulse)

	MCFG_CPU_ADD("rozcpu", Z80,XTAL1/4)         /* ? MHz */
	MCFG_CPU_PROGRAM_MAP(divebomb_rozcpu_map)
	MCFG_CPU_IO_MAP(divebomb_rozcpu_iomap)
	MCFG_CPU_VBLANK_INT_DRIVER("screen", divebomb_state,  nmi_line_pulse)

	MCFG_QUANTUM_PERFECT_CPU("fgcpu")

	MCFG_DEVICE_ADD("k051316_1", K051316, 0)
	MCFG_GFX_PALETTE("palette")
	MCFG_K051316_BPP(8)
	MCFG_K051316_WRAP(0)
	MCFG_K051316_OFFSETS(-88, -16)
	MCFG_K051316_CB(divebomb_state, zoom_callback_1)

	MCFG_DEVICE_ADD("k051316_2", K051316, 0)
	MCFG_GFX_PALETTE("palette")
	MCFG_K051316_BPP(8)
	MCFG_K051316_WRAP(0)
	MCFG_K051316_OFFSETS(-88, -16)
	MCFG_K051316_CB(divebomb_state, zoom_callback_2)



	/* video hardware */
	MCFG_SCREEN_ADD("screen", RASTER)
	MCFG_SCREEN_REFRESH_RATE(60)
	MCFG_SCREEN_VBLANK_TIME(ATTOSECONDS_IN_USEC(0))
	MCFG_SCREEN_SIZE(256, 256)
	MCFG_SCREEN_VISIBLE_AREA(0, 256-1, 0, 256-1-32)
	MCFG_SCREEN_UPDATE_DRIVER(divebomb_state, screen_update_divebomb)
	MCFG_SCREEN_PALETTE("palette")

	MCFG_GFXDECODE_ADD("gfxdecode", "palette", divebomb)
	MCFG_PALETTE_ADD("palette", 0x400+0x400+0x400+0x100)


	MCFG_PALETTE_INIT_OWNER(divebomb_state, divebomb)

	/* sound hardware */
	MCFG_SPEAKER_STANDARD_MONO("mono")

	MCFG_SOUND_ADD("sn0", SN76489, XTAL1/8) // ??
	MCFG_SOUND_ROUTE(ALL_OUTPUTS, "mono", 0.15)
	MCFG_SOUND_ADD("sn1", SN76489, XTAL1/8) // ??
	MCFG_SOUND_ROUTE(ALL_OUTPUTS, "mono", 0.15)
	MCFG_SOUND_ADD("sn2", SN76489, XTAL1/8) // ??
	MCFG_SOUND_ROUTE(ALL_OUTPUTS, "mono", 0.15)
	MCFG_SOUND_ADD("sn3", SN76489, XTAL1/8) // ??
	MCFG_SOUND_ROUTE(ALL_OUTPUTS, "mono", 0.15)
	MCFG_SOUND_ADD("sn4", SN76489, XTAL1/8) // ??
	MCFG_SOUND_ROUTE(ALL_OUTPUTS, "mono", 0.15)
	MCFG_SOUND_ADD("sn5", SN76489, XTAL1/8) // ??
	MCFG_SOUND_ROUTE(ALL_OUTPUTS, "mono", 0.15)



MACHINE_CONFIG_END



ROM_START( divebomb )
	ROM_REGION( 0x08000, "fgcpu", 0 )
	ROM_LOAD( "U20.27256", 0x00000, 0x08000, CRC(89a30c82) SHA1(ecb3dcc00192b9646dab55ec6c102ab1d2e403e8) )
	
	ROM_REGION( 0x08000, "spritecpu", 0 )
	ROM_LOAD( "U21.27256", 0x00000, 0x08000, CRC(3896d3f6) SHA1(86d10c907bf00977656d8fa426b2f1ac211acdb3) )

	ROM_REGION( 0x08000, "rozcpu", 0 )
	ROM_LOAD( "U19.27256", 0x00000, 0x08000, CRC(16e26fb9) SHA1(9602c79f947f5267e5f4f4f8dedc4697861a519b) )

	ROM_REGION( 0x40000, "rozcpudata", 0 ) // on sub-board
	ROM_LOAD( "U9.27512",  0x00000, 0x10000, CRC(c842f831) SHA1(badb38908a285e54b85e369342fd6ecb8fd68bbb) )
	ROM_LOAD( "U10.27512", 0x10000, 0x10000, CRC(c77f3574) SHA1(06b7576d5949906ee3209c011cd30e7066bb20cc) )
	ROM_LOAD( "U11.27512", 0x20000, 0x10000, CRC(8d46be7d) SHA1(7751df1f39b208169f04a5b904cb63e9fb53bba8) )
	// u12 not populated

	ROM_REGION( 0x10000, "gfx1", 0 )
	ROM_LOAD16_BYTE( "U22.27256", 0x00000, 0x08000, CRC(f816f9c5) SHA1(b8e136463a1b4c81960c6b7350472d82af0fb1fb) )
	ROM_LOAD16_BYTE( "U23.27256", 0x00001, 0x08000, CRC(d2600570) SHA1(a7f7e182670e7b95321c4ec8278ce915bbe2b5ca) )

	ROM_REGION( 0x80000, "gfx2", 0 )
	ROM_LOAD32_BYTE( "U15.27C100", 0x00000, 0x20000, CRC(ccba7fa0) SHA1(5eb4c1e458e7810e0f9db92946474d6da65f1a1b) )
	ROM_LOAD32_BYTE( "U16.27C100", 0x00001, 0x20000, CRC(16891fef) SHA1(a4723958509bccc73138306e58c355325ec342a3) )
	ROM_LOAD32_BYTE( "U17.27C100", 0x00002, 0x20000, CRC(f4cbc97f) SHA1(1e13bc18db128575ca8e6998e9dd6f7dc37a99b8) )
	ROM_LOAD32_BYTE( "U18.27C100", 0x00003, 0x20000, CRC(91ab9d89) SHA1(98454df4638bb831babb1796b095169851e6cf40) )

	ROM_REGION( 0x30000, "k051316_1", ROMREGION_INVERT ) // on sub-board
	ROM_LOAD( "U1.27512", 0x00000, 0x10000, CRC(99af1e18) SHA1(5d63130313fdd85c58e1d6b59e42b75a15328a6b) )
	ROM_LOAD( "U2.27512", 0x10000, 0x10000, CRC(99c8d516) SHA1(6205907bd526181542f4d58d442667595aec9730) )
	ROM_LOAD( "U3.27512", 0x20000, 0x10000, CRC(5ab4af3c) SHA1(ab8632e37a42f2f0db9b22c8577c4f09718ccc7c) )
	// u4 not populated

	ROM_REGION( 0x40000, "k051316_2", ROMREGION_INVERT ) // on sub-board
	ROM_LOAD( "U5.27512", 0x00000, 0x10000, CRC(6726d022) SHA1(314fcec87f3bf335356b24224158233d91012675) )
	ROM_LOAD( "U6.27512", 0x10000, 0x10000, CRC(756b8a12) SHA1(2869c18ef1592d00dbc340facac2002d21adf6bc) )
	ROM_LOAD( "U7.27512", 0x20000, 0x10000, CRC(01c07d84) SHA1(45e05f15e94a32adbd488a4f77a9619e7e6b63f3) )
	ROM_LOAD( "U8.27512", 0x30000, 0x10000, CRC(5b9e7caa) SHA1(85510c0b861bad6a1411afc1d628bc7c448c9fef) )

	// there are 12 PROMS, 4x banks of 3
	ROM_REGION( 0xc00, "fg_proms", 0 ) // text layer palette
	ROM_LOAD( "u35.MB7122.bin", 0x800, 0x400, CRC(e890259d) SHA1(f96e00da6de3b868e50f2347b83f8e5727b85d9b) )
	ROM_LOAD( "u36.MB7122.bin", 0x400, 0x400, CRC(0c1ecdb5) SHA1(4ba9e283c747d6a760fe2e0a8b2bfdfcc55a3969) )
	ROM_LOAD( "u37.MB7122.bin", 0x000, 0x400, CRC(55c17465) SHA1(153e99a09604467ddd8c641d60c5f1b8d9a205b4) )

	ROM_REGION( 0xc00, "k051316_2_pr", 0 ) // back roz ( k051316_2 )
	// 4 valid palettes but our code is only using 1
	ROM_LOAD( "u34.MB7122.bin", 0x000, 0x400, CRC(e0e2d93b) SHA1(9043929520abde15727e9333c153cf97104c9003) )
	ROM_LOAD( "u33.MB7122.bin", 0x400, 0x400, CRC(4df75f4f) SHA1(0157d6e268cdd797622c712648eb2e88214b12d9) )
	ROM_LOAD( "u32.MB7122.bin", 0x800, 0x400, CRC(e2e4b443) SHA1(e97a4e2988e29f992c5dec6f817a783b14535742) )

	ROM_REGION( 0xc00, "k051316_1_pr", 0 ) // front roz ( k051316_1 )
	// looks like banks 0,1,2 are valid
	ROM_LOAD( "u31.MB7122.bin", 0x000, 0x400, CRC(75cf5f3d) SHA1(ec02b99ab65e561596b773918569b28313f21835) ) // the last 0x100 block is invaldid (copy of data in u33)
	ROM_LOAD( "u30.MB7122.bin", 0x400, 0x400, CRC(0aeb1a88) SHA1(7b00e85ced210a5f7dfc100c15baa9e1735c7c80) ) // the last 0x100 block is empty
	ROM_LOAD( "u29.MB7122.bin", 0x800, 0x400, CRC(8b3d60d2) SHA1(a9ac4d3dd5e72522717dd18c32d3f88b75bb077c) ) // the last 0x100 block is invaldid (copy of data in u33)

	ROM_REGION( 0x300, "spr_proms", 0 ) // sprite layer palette
	ROM_LOAD( "u83.MB7114.bin", 0x000, 0x100, CRC(d216110d) SHA1(3de935dbf876f82864b8b69049c8681101619411) )
	ROM_LOAD( "u82.MB7114.bin", 0x100, 0x100, CRC(52637774) SHA1(919727e337a716dbd18b51e26e45025b82aeba79) )
	ROM_LOAD( "u81.MB7114.bin", 0x200, 0x100, CRC(c59b0857) SHA1(aea4cb8d1ab59b54f90edb96d4ac42b1dd6bdcbe) )
ROM_END 


// the World release was to be called 'Gaia' (based on a flyer)
GAME( 1989, divebomb,  0,    divebomb, divebomb, driver_device,  0, ROT270, "Konami", "Kyuukoukabakugekitai - Dive Bomber Squad (Japan, prototype)", MACHINE_IMPERFECT_GRAPHICS )

   

